import fs, { createReadStream } from 'fs';
import path from 'path';
import * as tar from 'tar';
import glob from 'fast-glob';
import { program, Command } from '@/program.ts';
import { getBaseURL, getConfig } from '@/module/index.ts';
import { fileIsExist } from '@/uitls/file.ts';
import ignore from 'ignore';
import FormData from 'form-data';

// 查找文件（忽略大小写）
async function findFileInsensitive(targetFile: string): Promise<string | null> {
  const files = fs.readdirSync('.');
  const matchedFile = files.find((file) => file.toLowerCase() === targetFile.toLowerCase());
  return matchedFile || null;
}

// 递归收集文件信息
async function collectFileInfo(filePath: string, baseDir = '.'): Promise<any[]> {
  const stats = fs.statSync(filePath);
  const relativePath = path.relative(baseDir, filePath);

  if (stats.isFile()) {
    return [{ path: relativePath, size: stats.size }];
  }

  if (stats.isDirectory()) {
    const files = fs.readdirSync(filePath);
    const results = await Promise.all(files.map((file) => collectFileInfo(path.join(filePath, file), baseDir)));
    return results.flat();
  }

  return [];
}
// 解析 .npmignore 文件
async function loadNpmIgnore(cwd: string): Promise<ignore.Ignore> {
  const npmIgnorePath = path.join(cwd, '.npmignore');
  const ig = ignore.default();

  try {
    const content = fs.readFileSync(npmIgnorePath, 'utf-8');
    ig.add(content);
  } catch (err) {
    console.warn('.npmignore not found, using default ignore rules');
    // 如果没有 .npmignore 文件，使用默认规则
    ig.add(['node_modules', '.git']);
  }

  return ig;
}
// 获取文件列表，兼容 .npmignore
async function getFiles(cwd: string, patterns: string[]): Promise<string[]> {
  const ig = await loadNpmIgnore(cwd);

  // 使用 fast-glob 匹配文件
  const allFiles = await glob(patterns, {
    cwd,
    dot: true, // 包括隐藏文件
    onlyFiles: false, // 包括目录
    followSymbolicLinks: true,
  });

  // 过滤忽略的文件
  const filteredFiles = allFiles.filter((file) => !ig.ignores(file));
  return filteredFiles;
}

export const pack = async () => {
  const cwd = process.cwd();
  const collection: Record<string, any> = {};
  const packageJsonPath = path.join(cwd, 'package.json');
  if (!fileIsExist(packageJsonPath)) {
    console.error('package.json not found');
    return;
  }

  let packageJson;
  try {
    const packageContent = fs.readFileSync(packageJsonPath, 'utf-8');
    packageJson = JSON.parse(packageContent);
  } catch (error) {
    console.error('Invalid package.json:', error);
    return;
  }

  let outputFileName = `${packageJson.name}-${packageJson.version}.tgz`
    .replace('@', '') // 替换特殊字符 @
    .replace(/[\/\\:*?"<>|]/g, '-'); // 替换特殊字符

  const outputFilePath = path.join(cwd, outputFileName);

  // 从 package.json 的 files 字段收集文件
  const filesToInclude = packageJson.files
    ? await glob(packageJson.files, {
        cwd: cwd,
        dot: true, // 包括隐藏文件
        onlyFiles: false, // 包括目录
        followSymbolicLinks: true, // 处理符号链接
      })
    : [];
  // 确保 README.md 和 dist 存在（忽略大小写检测 README.md）
  const readmeFile = await findFileInsensitive('README.md');
  if (readmeFile && !filesToInclude.includes(readmeFile)) {
    filesToInclude.push(readmeFile);
  }
  const packageFile = await findFileInsensitive('package.json');
  if (packageFile && !filesToInclude.includes(packageFile)) {
    filesToInclude.push(packageFile);
  }
  const allFiles = (await Promise.all(filesToInclude.map((file) => collectFileInfo(file)))).flat();

  // 输出文件详细信息
  console.log('Tarball Contents:');
  allFiles.forEach((file) => {
    console.log(`${file.size}B  ${file.path}`);
  });
  const totalSize = allFiles.reduce((sum, file) => sum + file.size, 0);
  const packageSize = (totalSize / 1024).toFixed(2) + ' kB';

  collection.files = allFiles;
  collection.packageJson = packageJson;
  collection.totalSize = totalSize;
  collection.tags = packageJson.app?.tags || packageJson.keywords || [];

  console.log('\nTarball Details');
  console.log(`name:    ${packageJson.name}`);
  console.log(`version: ${packageJson.version}`);
  console.log(`filename: ${outputFileName}`);
  console.log(`package size: ${packageSize}`);
  console.log(`total files: ${allFiles.length}`);
  console.log(`Created package: ${outputFileName}`);
  try {
    await tar.c(
      {
        gzip: true,
        file: outputFilePath,
        cwd: cwd,
      },
      filesToInclude,
    );
  } catch (error) {
    console.error('Error creating tarball:', error);
  }
  const readme = await findFileInsensitive('README.md');
  if (readme) {
    const readmeContent = fs.readFileSync(readme, 'utf-8');
    collection.readme = readmeContent;
  }
  return { collection, outputFilePath };
};
export const packByIgnore = async () => {
  let collection: Record<string, any> = {};
  const cwd = process.cwd();
  const patterns = ['**/*']; // 匹配所有文件
  const packageJsonPath = path.join(cwd, 'package.json');
  let packageJson;
  try {
    const packageContent = fs.readFileSync(packageJsonPath, 'utf-8');
    packageJson = JSON.parse(packageContent);
  } catch (error) {
    console.error('Invalid package.json:', error);
    return;
  }

  let outputFileName = `${packageJson.name}-${packageJson.version}.tgz`
    .replace('@', '') // 替换特殊字符 @
    .replace(/[\/\\:*?"<>|]/g, '-'); // 替换特殊字符

  const outputFilePath = path.join(cwd, outputFileName);

  // 获取符合条件的文件列表
  const files = await getFiles(cwd, patterns);

  console.log('Files to include in the package:');
  // files 获取 size 和 path
  const filesInfo = await Promise.all(files.map((file) => collectFileInfo(file)));
  const allFiles = filesInfo.flat();
  allFiles.forEach((file) => {
    console.log(`${file.size}B  ${file.path}`);
  });

  const totalSize = allFiles.reduce((sum, file) => sum + file.size, 0);
  const packageSize = (totalSize / 1024).toFixed(2) + ' kB';

  collection.files = allFiles;
  collection.packageJson = packageJson;
  collection.totalSize = totalSize;
  collection.tags = packageJson.app?.tags || packageJson.keywords || [];

  console.log('\nTarball Details');
  console.log(`package size: ${packageSize}`);
  console.log(`total files: ${allFiles.length}`);
  const filesToInclude = files.map((file) => path.relative(cwd, file));
  try {
    await tar.c(
      {
        gzip: true,
        file: outputFilePath,
        cwd: cwd,
      },
      filesToInclude,
    );
  } catch (error) {
    console.error('Error creating tarball:', error);
  }
  const readme = await findFileInsensitive('README.md');
  if (readme) {
    const readmeContent = fs.readFileSync(readme, 'utf-8');
    collection.readme = readmeContent;
  }
  return { collection, outputFilePath };
};
export const packLib = async (ignore: boolean = false) => {
  if (ignore) {
    return await packByIgnore();
  }
  return await pack();
};
export const unpackLib = async (filePath: string, cwd: string) => {
  try {
    await tar.x({
      file: filePath,
      cwd: cwd,
    });
  } catch (error) {
    console.error('Error extracting tarball:', error);
  }
};
const publishCommand = new Command('publish')
  .description('发布应用')
  .option('-k, --key <key>', '应用 key')
  .option('-v, --version <version>', '应用版本')
  .action(async (options) => {
    const { key, version } = options;
    const config = await getConfig();
    console.log('发布逻辑实现', { key, version, config });
  });
const uploadFiles = async (filePath: string, collection: any): Promise<any> => {
  const config = await getConfig();
  const form = new FormData();
  const filename = path.basename(filePath);
  try {
    console.log('upload file', filePath);
    form.append('file', createReadStream(filePath));
  } catch (error) {
    console.error('Error reading file:', error);
    return;
  }
  form.append('collection', JSON.stringify(collection));
  console.log('upload file', filename);
  return await new Promise((resolve) => {
    const _baseURL = getBaseURL();
    const url = new URL(_baseURL);
    console.log('upload url', url.hostname, url.protocol, url.port);
    form.submit(
      {
        path: '/api/micro-app/upload',
        host: url.hostname,
        protocol: url.protocol as any,
        port: url.port,
        method: 'POST',
        headers: {
          Authorization: 'Bearer ' + config.token,
          ...form.getHeaders(),
        },
      },
      (err, res) => {
        if (err) {
          console.error('Error uploading file:', err.message);
          return;
        }
        console.log('upload success');
        // 处理服务器响应
        let body = '';
        res.on('data', (chunk) => {
          body += chunk;
        });

        res.on('end', () => {
          try {
            const res = JSON.parse(body);
            console.log('upload success', res);
            resolve(res);
          } catch (e) {
            resolve({ code: 500, message: body });
          }
        });
      },
    );
  });
};
const packCommand = new Command('pack')
  .description('打包应用, 默认使用 package.json 中的 files 字段')
  .option('-i, --ignore', '使用 .npmignore 文件模式去忽略文件进行打包, 不需要package.json中的files字段')
  .option('-p, --publish', '打包并发布')
  .action(async (opts) => {
    let value: { collection: Record<string, any>; outputFilePath: string } = await packLib(opts.ignore);
    if (opts.publish && value?.collection) {
      console.log('\n\npublish');
      const { collection, outputFilePath } = value;
      await uploadFiles(outputFilePath, collection);
    }
    //
  });

program.addCommand(publishCommand);
program.addCommand(packCommand);
